from typing import Optional, List, Any
from qumeas.state_prep import StatePreparation

class qiskitStatePreparation(StatePreparation):
    """
    State preparation class using Qiskit
    """
    
    def __init__(self,
                 Nqubit:Optional[int] = None,
                 state_vector: Optional[List[float]] = None,
                 state_preparation_circuit: Optional['QuantumCircuit'] = None,
                 backend: Optional[Any] = None
                 ):
        """
        Initialize the qiskitStatePreparation class.

        Parameters
        ----------
        Nqubit : Optional[int]
            Number of qubits.
        state_vector : Optional[List[float]]
            State vector. If `state_preparation_circuit` is not provided, state_vector is used for initialization.
        state_preparation_circuit : Optional['QuantumCircuit']
            Qiskit QuantumCircuit for state preparation.
        backend : Optional[Any]
            Qiskit backend. Defaults to `statevector` in `AerSimulator`
        """
        super().__init__(Nqubit, state_vector, state_preparation_circuit, backend)

    def get_measurement_circuit(self, pauli_string: List[str]):
        """
        Generate a measurement circuit based on a Pauli string.

        Parameters
        ----------
        pauli_string : List[str]
            Pauli string for which expectation value is sought.

        Returns
        -------
        QuantumCircuit
            Quantum circuit with measurement gates
        """
        import numpy
        from qiskit import QuantumCircuit
        
        measurement_circuit = QuantumCircuit(self.Nqubit, self.Nqubit)
        for idx, i in enumerate(pauli_string):
            if i == 'X' or i == 1:
                measurement_circuit.h(idx)
            elif i == 'Y' or i == 2:
                measurement_circuit.h(idx)
                measurement_circuit.p(-numpy.pi / 2., idx)
                
        return measurement_circuit
        
    def get_circuit(self,
                    measurement_circuit: Optional['QuantumCircuit'] = None,
                    ):
        """
        Construct the quantum circuit for state preparation, including measurement gates.

        Parameters
        ----------
        measurement_circuit : Optional['QuantumCircuit']
            Qiskit QuantumCircuit object with measurement gates.

        Returns
        -------
        QuantumCircuit
            Quantum circuit (Qiskit QuantumCircuit) for state preparation and measurement.
        """
        from qiskit import QuantumCircuit

        qcircuit = QuantumCircuit(self.Nqubit, self.Nqubit)

        # Construct quantum circuit
        if self.state_vector is not None:
            qcircuit.initialize(self.state_vector,
                                reversed(range(self.Nqubit)))            
        elif self.state_preparation_circuit is not None:
            qcircuit = qcircuit.compose(self.state_preparation_circuit)

        # Add measurement circuit if provided
        if measurement_circuit is not None:
            qcircuit = qcircuit.compose(measurement_circuit)
        qcircuit.measure(reversed(range(self.Nqubit)), range(self.Nqubit))

        return qcircuit


    def measure(self,
                pauli_string: List[str],
                shots: Optional[int] = 1,
                seed: Optional[int] = None
                ):
        """
        Measure the quantum state to estimate expectation value of `pauli_string`

        Parameters
        ----------
        pauli_string : List[str]
            Pauli string to measure.
        shots : Optional[int], default=1
            Number of measurement shots. `RandomShadow` as well as `QCumulant` only supports one shot measurements for now.
        seed : Optional[int], default=None
            Seed for the simulator.

        Returns
        -------
        List[Tuple[List[int], int]]
            List of tuples containing classical measurement bits and their counts.
        """
        from qiskit import transpile
        from qiskit_aer import AerSimulator
        
        measurement_circuit = self.get_measurement_circuit(pauli_string)
        qcircuit = self.get_circuit(measurement_circuit = measurement_circuit)
        
        if self.backend is None:
            self.backend = AerSimulator(method = 'statevector', seed_simulator=seed)
            
        qcircuit = transpile(qcircuit, self.backend, optimization_level=0)
        res = self.backend.run(qcircuit, shots = shots).result()
        counts = res.get_counts()
        
        outbit_list = [
            ([1 if bit == '0' else -1 for bit in bitstring], count)
            for bitstring, count in counts.items()
        ]

        return outbit_list



from typing import Optional, List, Dict, Any, Union
from .helpers import *
from .pauli_container import PauliContainer
from .random_measurement import RandomShadow
from .qiskit_utils import Mole

class QCumulant:
    """
    Compute expectation values for Pauli strings using Cumulant expansion and non-crossing partitioning.
    """
    
    def __init__(self,
                 protocol: Optional[Union[RandomShadow, Mole]] = None,
                 PauliObj: Optional[PauliContainer] = None,
                 measure_basis: Optional[List[Union[str, List[int]]]] = None,
                 measure_outcome_bits: Optional[List[Union[str, List[int]]]] = None,
                 partitions: Optional[List[List[List[int]]]] = None):
        """
        Initialize QCumulant class.

        Parameters
        ----------
        protocol : Optional[Union[RandomShadow, Mole]]
            Provide either a `RandomShadow` or `Mole` object from qumeas.
        PauliObj : Optional[PauliContainer]
            If protocol is not provided, `PauliContainer` object can be passed in. Expectation values can be computed using only the state vector is outbits are not provided along with `PauliObj`.
        measure_basis: List[Union[str, List[int]]]
            Measurement basis. Either list of Pauli operators (strings) or list of list of int representing Paulis, with 1 = 'X', 2 = 'Y', and 3 = 'Z'.
        measure_outcome_bits: List[Union[str, List[int]]]
            Measurement outcome (classical bits). Either a list of string or list of list of int. Binary bits can be passed in as string (e.g. ['0011', '0010']. Also, supports a list, e.g. [[0, 0, 1, 1], [0, 0, 1, 0]] or in this format: [[1, 1, -1, -1],[1, 1, -1, 1]].
        partitions : Optional[List[List[List[int]]]]
            Provide your partitions of `protocol.PauliObj.pauli_list` or `PauliObj.pauli_list`    

        """
        
        # Initialize protocol attributes
        self._initialize_protocol_attributes(protocol, PauliObj)
        
        if measure_basis is not None and measure_outcome_bits is not None:
            measure_basis, measure_outcome_bits = process_measure_bits(measure_basis, measure_outcome_bits)
            self.outbits = [measure_basis, measure_outcome_bits]

        if self.PauliObj.state_vector is None and self.outbits is None:
            raise ValueError("Requires either a state vector or measurment classical bits in Qcumulant")

        self.partitions = partitions

    def _initialize_protocol_attributes(self, protocol, PauliObj):
        
        if isinstance(protocol, (RandomShadow, Mole)):
            self.PauliObj = protocol.PauliObj
        elif isinstance(PauliObj, PauliContainer):
            self.PauliObj = PauliObj
        else:
            raise TypeError(
                "Neither 'protocol.PauliObj' nor 'PauliObj' is an instance of PauliContainer."
                )

        if isinstance(protocol, Mole) or isinstance(PauliObj, PauliContainer):
            self._process_pauli_string()
        
        if isinstance(protocol, RandomShadow):
            self.outbits = protocol.outbits

    def _process_pauli_string(self):
        from .helpers import string2int

        if isinstance(self.PauliObj.pauli_list[0], str):
            self.PauliObj.pauli_list = [string2int(_) for _ in self.PauliObj.pauli_list]
        elif isinstance(self.PauliObj.pauli_list[0], list) and all(isinstance(_, int) for _ in self.PauliObj.pauli_list[0]):
            pass
        else:
            raise TypeError(
                "PauliObj.pauli_list elements must be either str or a list of int")

    def generate_partitions(self, max_size=4, num_threads=1):
        """
        Generate non-crossing partitions for Pauli strings in `PauliObj` (or `protocol.PauliObj`).

        Parameters
        ----------
        max_size : int, optional
            Maximum block size for non-crossing partitions (default is 4).
        num_threads : int, optional
            Number of threads to use for partition generation (default is 1).

        Raises
        ------
        ValueError
            If `PauliObj.pauli_list` is empty.
        """
        from qumeas.libmeas import generate_partition_non_crossing
        
        if not self.PauliObj.pauli_list:
            raise ValueError("pauli_list is required for generating partitions.")
        
        non_ilist_full = [
            [idx for idx, i in enumerate(ham_string) if i != 'I' and i != 0]
            for ham_string in self.PauliObj.pauli_list
        ]
        self.non_ilist_full = non_ilist_full
        
        # Generate partitionings & cumulants
        self.partitions = generate_partition_non_crossing(
            self.non_ilist_full, max_size, num_threads)
        
        self._generate_unique_partitions()


    def _generate_unique_partitions(self):
        """Get unique partitions from partitions of a list of Pauli strings"""
        
        self.cdict = {}
        self.clist = []

        for idx, ham_string in enumerate(self.PauliObj.pauli_list):
            
            if not self.non_ilist_full[idx]:
                continue
            
            for block_list in self.partitions[idx]:
                for blocks in block_list:

                    # Get a compressed key (string)
                    tkey = sblock2ndict(blocks, ham_string)
                    tstring = sblock2ncumu(blocks, ham_string)
                    if tkey not in self.cdict:
                        self.cdict[tkey] = 0.0
                        self.clist.append([tkey, blocks, tstring])
                 
    def compute_expectation_bits(self, return_Elist = False):
        """
        Compute expectation values for the cumulant expansion using quantum measurement bits. This method uses `parition_expectation_bits` function in `libmeas` that uses asynchronous tasks to parallelize each cumulant expectation computation. Each cumulant expectation computation is OpenMP parallelized for optimal performance.

        Parameters
        ----------
        return_Elist : bool, optional
            If True, return both expectation value and a list of expectation values for each Pauli string (default is False).

        Returns
        -------
        Union[float, Tuple[float, List[float]]]
            The computed expectation value and optionally the list of expectation values.
        """
        from qumeas.libmeas import partition_expectation_bits

        cdict = partition_expectation_bits(self.outbits[0], self.outbits[1], self.clist)
        
        return self._compute_expectation(cdict, return_Elist)

    def compute_expectation_state(self, return_Elist = False):
        """
        Compute expectation values for the cumulant expansion using the state vector in `PauliObj` (or `protocol.PauliObj`). This method uses `parition_expectation_state` function in `libmeas` that uses asynchronous tasks to parallelize each cumulant expectation computation. Each cumulant expectation computation is OpenMP parallelized for optimal performance.

        Parameters
        ----------
        return_Elist : bool, optional
            If True, return both expectation value and a list of expectation values for each Pauli string (default is False).

        Returns
        -------
        Union[float, Tuple[float, List[float]]]
            The computed expectation value and optionally the list of expectation values.
        """
        from qumeas.libmeas import partition_expectation_state
        
        cdict = partition_expectation_state(self.PauliObj.Nqubit, self.PauliObj.state_vector,
                                           self.clist)
        return self._compute_expectation(cdict, return_Elist)

    def _compute_expectation(self, cdict, return_Elist):
        
        expectation_list = self._get_expect_list(cdict)
        expectation_val = self._compute_expectation_tot(expectation_list)
        
        return (expectation_val, expectation_list) if return_Elist else expectation_val


    def _compute_expectation_tot(self, expectation_list):
        return sum(coeff * expectation_list[idx] for idx, coeff in enumerate(self.PauliObj.pauli_list_coeff))


    def _get_expect_list(self, cdict):
              
        expec_list = []
        
        for idx, ham_string in enumerate(self.PauliObj.pauli_list):
                    
            if not len(self.non_ilist_full[idx]):
                expec_list.append(1.0)
                continue
            
            sum_expect = 0.0                
            for block_list in self.partitions[idx]:
                
                expect_part = 1.0               
                for blocks in block_list:                    
                    tkey = sblock2ndict(blocks, ham_string)
                    expect_part *= cdict[tkey]                    
                sum_expect += expect_part
                
            expec_list.append(sum_expect)
        
        return expec_list



from typing import Optional, List, Any
from .pauli_container import PauliContainer
from .state_prep import StatePreparation

class RandomShadow:
    """
    Class for estimating expectation values of Pauli strings using classical shadow tomography (randomized measurement basis).

    Attributes
    ----------
    PauliObj : PauliContainer
        `PauliContainer` object holding Pauli strings, coefficients, state vector, and number of qubits.
    basis : Optional[List[List[int]]]
        Basis for measurements in classical shadow tomography.
    state_prep : Optional[StatePreparation]
        State preparation object, can be an instance of StatePreparation or qiskitStatePreparation.
    outbits : Optional[List[List[List[int]]]]
        Measurement output bits. Contains list of measurement basis and another list of the corresponding classical bits. In the measurement basis, X,Y,Z are represented as integers 1,2,3 respectively.
    """
    
    def __init__(self,
                 mole: Optional[Any] = None,
                 PauliObj: Optional[PauliContainer] = None,
                 state_prep: Optional[StatePreparation] = None,
                 basis: Optional[List[List[int]]] = None
                 ):
        """
        Initialize RandomShadow class.

        Parameters
        ----------
        mole : Optional[Any]
            `Mole` object that contains a `PauliContainer` instance.
        PauliObj : Optional[PauliContainer]
            `PauliContainer` object holding Pauli strings, coefficients, state vector, and number of qubits.
        state_prep : Optional[StatePreparation]
            State preparation object, can be an instance of StatePreparation or qiskitStatePreparation.
        basis : Optional[List[List[int]]]
            Basis for measurements in classical shadow tomography.

        Raises
        ------
        TypeError
            If neither `mole.PauliObj` nor `PauliObj` is a PauliContainer instance.
        """
        
        if hasattr(mole, 'PauliObj') and isinstance(mole.PauliObj, PauliContainer):
            self.PauliObj = mole.PauliObj
        elif isinstance(PauliObj, PauliContainer):
            self.PauliObj = PauliObj
        else:
            raise TypeError(
                "Neither 'mole.PauliObj' nor 'PauliObj' is an instance of PauliContainer.")

        self._process_pauli_string() 
        self.basis = basis            
        self.state_prep = state_prep
        self.outbits = None

    def _process_pauli_string(self):
        from .helpers import string2int
        
        if isinstance(self.PauliObj.pauli_list[0], str):
            self.PauliObj.pauli_list = [string2int(_) for _ in self.PauliObj.pauli_list]
        elif isinstance(self.PauliObj.pauli_list[0], list) and all(isinstance(_, int) for _ in self.PauliObj.pauli_list[0]):
            pass
        else:
            raise TypeError(
                "PauliObj.pauli_list elements must be either str or a list of int")
        
        
    def get_basis(self,
                  M: int,
                  seed: Optional[int] = None):
        """
        Generate a random basis for measurements.

        Parameters
        ----------
        M : int
            Number of random basis to generate.
        seed : Optional[int], default=None
            Random seed for reproducibility.
        """
        
        self.basis = self._generate_random_basis(M, seed = seed)
        

    def _generate_random_basis(self, M:int,
                               seed: Optional[int] = None):
        import numpy
        
        basis = []
        for i in range(M):            
            if seed is not None:
                numpy.random.seed(seed+i)
            basis.append([int(numpy.random.choice([1,2,3])) for _ in range(self.PauliObj.Nqubit)])
        
        return basis        

    def measure(self,
                shots: int = 1,
                M: int = 100,
                seed: Optional[int] = None,
                nproc: Optional[int] = 1
                ):

        """
        Perform measurements using the state preparation object (`state_prep`).

        Parameters
        ----------
        shots : int, default=1
            Number of measurement shots. Currently, only supports for one shot measurements.
        M : int, default=100
            Number of random basis in classical shadow tomography
        seed : Optional[int], default=None
            Random seed for reproducibility.
        nproc : Optional[int], default=1
            Number of processes for parallelization accross measurement bases. Each measurement basis can be processed in parallel across multiple processes, with each measurement further parallelized using OpenMP for optimal performance.

        Returns
        -------
        float
            Computed expectation value.
        """
        
        self._prepare_basis_and_state(M, seed)

        if nproc == 1:
            bitplist, bitoutlist = self._serial_measurement(shots, seed=seed)
        else:
            bitplist, bitoutlist = self._parallel_measurement(shots, nproc, seed=seed)
        
        self.outbits = [bitplist, bitoutlist]#, bitcount]

        expectation_val = self.compute_expectation(bitplist, bitoutlist)#, bitcount)
        return expectation_val
    
    def _prepare_basis_and_state(self, M: int,
                                 seed: Optional[int] = None):
        """Generate basis and set state_prep to default qiskitStatePreparation if not set."""
        from qumeas.qiskit_utils import qiskitStatePreparation
        
        if not self.basis:
            self.get_basis(M, seed)
            
        if not self.state_prep and self.PauliObj.Nqubit:
            self.state_prep = qiskitStatePreparation(Nqubit=self.PauliObj.Nqubit, state_vector=self.PauliObj.state_vector)
        

    def _serial_measurement(self, shots: int, seed: Optional[int] = None):
        """Perform measurement in serial mode."""
        
        bitplist, bitoutlist, bitcount = [], [], []
        
        for pbasis in self.basis:
            noutbit = self.state_prep.measure(pbasis, shots=shots, seed = seed)
            for outbit_ in noutbit:
                bitplist.append(pbasis)
                bitoutlist.append(outbit_[0])
                #bitcount.append(outbit_[1])
                
        return bitplist, bitoutlist#, bitcount

    def _parallel_measurement(self, shots: int, nproc: int, seed: Optional[int] = None):
        """Perform measurement in parallel using multiprocessing."""
        from multiprocessing import Pool
        
        results = []
        pool = Pool(nproc)
        outresult = []
        for measure_ in range(len(self.basis)):
            result = pool.apply_async(self.state_prep.measure, [self.basis[measure_], shots, seed])
            results.append(result)
    
        [outresult.append(result.get()) for result in results]
        pool.close()
        bitplist, bitoutlist, bitcount = [], [], []
        
        for pidx, pbasis in enumerate(self.basis):
            for outbit_ in outresult[pidx]:
                bitplist.append(pbasis)
                bitoutlist.append(outbit_[0])
                #bitcount.append(outbit_[1])
                
        return bitplist, bitoutlist#, bitcount

    def compute_expectation(self, measure_basis, measure_outcome_bits):
        """Compute the expectation value from measurement bits.

        Parameters
        ----------
        measure_basis: List[Union[str, List[int]]]
            Measurement basis. Either list of Pauli operators (strings) or list of list of int representing Paulis, with 1 = 'X', 2 = 'Y', and 3 = 'Z'.
        measure_outcome_bits: List[Union[str, List[int]]]
            Measurement outcome (classical bits). Either a list of string or list of list of int. Binary bits can be passed in as string (e.g. ['0011', '0010']. Also, supports a list, e.g. [[0, 0, 1, 1], [0, 0, 1, 0]] or in this format: [[1, 1, -1, -1],[1, 1, -1, 1]].

        Returns
        -------
        float
            Total expectation value.
        """
        from qumeas.libmeas import compute_expectations_basis
        from .helpers import process_measure_bits

        bitplist, bitoutlist = process_measure_bits(measure_basis, measure_outcome_bits)
        expectation_list = compute_expectations_basis(
            bitplist, bitoutlist, self.PauliObj.pauli_list, self.PauliObj.Nqubit
        )
        return sum(coeff * expectation_list[pdx] for pdx, coeff in enumerate(self.PauliObj.pauli_list_coeff))




def string2int(pstring):
    tmp_ = []
    for i in pstring:
        if i=='X':
            tmp_.append(1)
        elif i=='Y':
            tmp_.append(2)
        elif i=='Z':
            tmp_.append(3)
        else:
            tmp_.append(0)
    return tmp_


def sblock2ndict(block, Hstring):
    tstring = ""
    for _ in block:
        tstring += str(_)+' '+str(Hstring[_])
        
    return tstring

def sblock2nqubit(block, Hstring):
    tstring = [0 for i in Hstring]
    for j_ in block:
        tstring[j_] = Hstring[j_]
    return tstring

def sblock2ncumu(block, Hstring):
    tstring = []
    for j_ in block:
        tstring.append(Hstring[j_])
    return tstring

def process_measure_bits(bitplist, bitolist):
    from typing import Optional, List, Any
    
    if isinstance(bitplist, List) and isinstance(bitplist[0], str):
        plist_int = [string2int(_) for _ in bitplist]
    elif isinstance(bitplist[0], List) and all(isinstance(bitplist[0][_], int) for _ in bitplist[0]):
        pass
    else:
        raise ValueError("Unrecognized measurement basis format")

    if isinstance(bitolist, List) and isinstance(bitolist[0], str):
        olist_int = [[1 if _ == '0' else -1 for _ in bits] for bits in bitolist]
    elif isinstance(bitolist[0], List) and all(isinstance(bitolist[0][_], int) for _ in bitolist[0]):
        
        if check_binary(bitolist):
            olist_int = [[1 if _ == 0 else -1 for _ in bits] for bits in bitolist]
        else:
            pass
        
    try:
        return plist_int, olist_int
    except:
        return bitplist, bitolist

def check_binary(list_of_list):
    for i in list_of_list:
        for j in i:
            if j == 0:
                return True
            elif j == -1:
                return False
    return False



from dataclasses import dataclass, field
from typing import List, Optional

@dataclass
class PauliContainer:
    """
    Container for managing Pauli string data and state vector.

    Attributes
    ----------
    Nqubit : Optional[int]
        Number of qubits.
    pauli_list : List[str]
        List of Pauli strings.
    pauli_list_coeff : List[float]
        Coefficients corresponding to each Pauli string in `pauli_list`.
    state_vector : Optional[List[float]]
        State vector, if available.
    total_energy : Optional[float]
        Total ground state energy.
    """
    Nqubit: Optional[int] = None
    pauli_list: List[str] = field(default_factory=list)
    pauli_list_coeff: List[float] = field(default_factory=list)
    state_vector: Optional[List[float]] = None
    total_energy: Optional[float] = None



from skbuild import setup
from setuptools import find_packages
import os, sys


# Read VERSION and write python/qumeas/_version.py
def read_version():
    here = os.path.abspath(os.path.dirname(__file__))
    version_file = os.path.join(here, "VERSION")
    with open(version_file, "r", encoding="utf-8") as vf:
        return vf.read().strip()

version = read_version()
version_py = os.path.join("python", "qumeas", "_version.py")
with open(version_py, "w", encoding="utf-8") as vp:
    vp.write(f"__version__ = '{version}'\n")


# Read the long description from README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Define the setup configuration
setup(
    name="qumeas",
    version=version,
    author="Oinam Romesh Meitei",
    author_email="oinam.meitei@fau.de",
    description="A high-performance, multi-threaded, quantum computing library for Pauli measurements.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/oimeitei/qumeas",
    packages=find_packages(where='python'),
    package_dir={'': 'python'},
    cmake_install_dir='qumeas',
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",  
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    python_requires=">=3.8",
    install_requires=[
        "numpy",
        "scipy"
    ],
    extras_require={
        "qiskit": ["qiskit", "qiskit_nature", "qiskit_aer"],
    },
)




class StatePreparation:
    """
    Base class for quantum state preparation.
    
    """
    def __init__(self,
                 Nqubit = None,
                 state_vector = None,
                 state_preparation_circuit = None,
                 backend = None
                 ):
        """
        Initialization for the StatePreparation class.

        Parameters
        ----------
        Nqubit : Optional[int]
            Number of qubits.
        state_vector : Optional[List[float]]
            State vector. If `state_preparation_circuit` is not provided, state_vector is used for initialization.
        state_preparation_circuit : Optional[Any]
            Quantum circuit for state preparation.
        backend : Optional[Any]
            Backend for state preparation.
        """

        self.Nqubit = Nqubit
        self.state_vector = state_vector
        self.state_preparation_circuit = state_preparation_circuit
        self.backend = backend

        if self.state_vector is None and self.state_preperation_circuit is None:
            raise ValueError("Provide either 'state_vector' or 'state_preparation_circuit")

    def get_circuit(self):
        """
        Get the state preparation circuit.
        """
        pass

    def measure(self):
        """
        Measure the prepared quantum state
        """
        pass



from typing import Optional, List, Any
from qumeas.pauli_container import PauliContainer

class Mole:
    """
    A class representing a molecular system using Qiskit functionalities.

    Attributes
    ----------
    PauliObj : PauliContainer
        Container for Pauli strings, coefficients, state vector, and number of qubits.
    Hamiltonian : sparse matrix
        The Hamiltonian of the molecule, mapped to a qubit basis.
    """

    def __init__(self,
                 atom: Optional[str] = None,
                 molecule: Optional[str] = None,
                 basis: str = 'sto-3g',
                 charge: int = 0,
                 spin: int = 0,
                 qubit_mapper: str = 'jw',
                 frozen: bool = True,
                 norb: Optional[int] = None,
                 nelec: Optional[int] = None,
                 orb_indices: Optional[List[int]] = None
                 ):
        """
        Initialize the Mole class with molecular properties and quantum simulation parameters. Uses Qiskit's PySCFDriver.

        Parameters
        ----------
        atom : Optional[str]
            Cartesian coordinates for the geometry as a string, or `None` if `molecule` is provided.
        molecule : Optional[str]
            Name of the molecule (e.g., 'h2', 'lih'); overrides `atom` if specified. Uses predefined geometry of the molecule.
        basis : str, default='sto-3g'
            Quantum chemitry basis set.
        charge : int, default=0
            Charge.
        spin : int, default=0
            Molecular spin. 
        qubit_mapper : str, default='jw'
            Qubit mapping method ('jw' for Jordan-Wigner, 'bk' for Bravyi-Kitaev).
        frozen : bool, default=True
            Whether to freeze core orbitals.
        norb : Optional[int]
            Number of active spatial orbitals.
        nelec : Optional[int]
            Number of active electrons.
        orb_indices : Optional[List[int]]
            Indices of orbitals to include in the active space.

        Raises
        ------
        ValueError
            If neither `atom` nor `molecule` is provided, or if `qubit_mapper` is not 'jw' or 'bk'.
        """
        
        self.PauliObj = PauliContainer()
        
        if molecule:
            atom = self._get_atom(molecule)
        elif atom is None:
            raise ValueError("Either 'atom' or 'molecule' must be provided.")
        
        self.qubit_mapper = qubit_mapper.lower()
        if self.qubit_mapper not in ['jw', 'bk']:
            raise ValueError("qubit_mapper must be 'jw' (Jordan-Wigner) or 'bk' (Bravyi-Kitaev).")
        
        self.basis = basis
        self.charge = charge
        self.spin = spin
        self.frozen = frozen
        self.norb = norb
        self.nelec = nelec
        self.orb_indices = orb_indices

        self._initialize_driver(atom)

    def _initialize_driver(self, atom: str):
        from qiskit_nature.second_q.drivers import PySCFDriver
        from qiskit_nature.second_q.transformers import FreezeCoreTransformer, ActiveSpaceTransformer
        from qiskit_nature.units import DistanceUnit

        # Using pyscf
        driver = PySCFDriver(atom=atom,
                             unit=DistanceUnit.ANGSTROM,
                             charge=self.charge,
                             spin=self.spin,
                             basis=self.basis)
        problem = driver.run()
        
        if self.frozen:
            fc_transformer = FreezeCoreTransformer()
            problem = fc_transformer.transform(problem)
            
        if self.norb is not None and self.nelec is not None:
            if self.orb_indices is None:
                as_transformer = ActiveSpaceTransformer(self.nelec, self.norb)
            else:
                as_transformer = ActiveSpaceTransformer(self.nelec, self.norb, self.orb_indices)
            problem = as_transformer.transform(problem)
            
        self._map_hamiltonian(problem)


    def _map_hamiltonian(self, problem):
        
        hamiltonian = problem.hamiltonian
        second_q_op = hamiltonian.second_q_op()

        mapper = self._get_mapper(self.qubit_mapper)
        mapped_Ham = mapper.map(second_q_op)
        
        self.PauliObj.pauli_list = mapped_Ham.paulis.to_labels()
        self.PauliObj.pauli_list_coeff = mapped_Ham.coeffs.real
        self.PauliObj.Nqubit = len(self.PauliObj.pauli_list[0])

        self.Hamiltonian = mapped_Ham.to_matrix(sparse = True)
        self._get_state_vector()

        
    def _get_state_vector(self):
        import scipy
        
        eigE_, eigV_ = scipy.sparse.linalg.eigsh(self.Hamiltonian, which='SA', k=1)
        self.PauliObj.total_energy = eigE_[0]
        self.PauliObj.state_vector = eigV_[:,0]
        

    def _get_mapper(self, mapper_type: str):
        from qiskit_nature.second_q.mappers import BravyiKitaevMapper, JordanWignerMapper
        
        mapper_type = mapper_type.lower()        
        if mapper_type == 'bk':
            return BravyiKitaevMapper()
        elif mapper_type == 'jw':
            return JordanWignerMapper()
        else:
            raise ValueError(f"Unsupported qubit mapper: {mapper_type}. Use 'jw' or 'bk'.")
        

    def _get_atom(self, molecule: str):

        molecule_dict = {
            'h2': "H 0 0 0; H 0 0 0.735",
            'lih': "Li 0 0 0; H 0 0 1.45",
            'h2o': 'O 0.0 0.0 0.0; H 0.757 0.586 0.0; H -0.757 0.586 0.0',
            'beh2': 'H 0.0 0.0 -1.4; Be 0.0 0.0 0.0; H 0.0 0.0 1.4',
            'n2': "N 0 0 0; N 0 0 1.1",
            'o2': "O 0 0 0; O 0 0 1.16",
            'nh3': 'H 0.0 0.0 0.0; N 1.0296 0.0 0.0; H 1.0296 1.0296 0.0; H 1.0296 1.0296 -1.0296',
            'co': 'C 0.0 0.0 0.648; O 0.0 0.0 -0.486',
            'ch4': 'C 0.0 0.0 0.0; H 0.885 0.0 0.626; H 0.0 -0.885 -0.626; H 0.0 0.885 -0.626; H -0.885 0.0 0.626',
            'c6h6': 'C 0.0 1.3968 0.0; C 0.0 -1.3968 0.0; C 1.2097 0.6984 0.0; C -1.2097 -0.6984 0.0; C -1.2097 0.6984 0.0; C 1.2097 -0.6984 0.0; H 0.0 2.4842 0.0; H 2.1514 1.2421 0.0; H -2.1514 -1.2421 0.0; H -2.1514 1.2421 0.0; H 2.1514 -1.2421 0.0; H 0.0 -2.4842 0.0',
            'b2': 'B 0.0 0.0 0.0; B 0.0 0.0 1.59',
            'nh4+': 'N -1.1255 0.6901 -0.1099; H -0.2133 1.1642 0.07; H -1.2866 0.6009 0.9425; H -1.9221 0.0372 -0.2638; H -0.3508 -0.3076 -0.3863'
        }

        if molecule in molecule_dict:
            return molecule_dict[molecule]
        else:
            raise ValueError(f"Molecule '{molecule}' is not defined in the molecular dictionary.")
